
:- multifile syntax/3.
:- discontiguous syntax/3.
:- dynamic syntax/3.
:- dynamic saved/2.
:- dynamic lastSymbol/1.
:- dynamic loadingFrom/1.


:-op(150,xfy,@).
:-op(150,xfy,^^).
:-op(1200,xfx,'::=').
:- dynamic (::=)/2.
:- discontiguous (::=)/2.

saveUp(File) :-
   consult(File),
   retract('::='(A,B)),
   assertz(syntax(A,B,File)),
   fail.
saveUp(_) :-
   abolish(('::='),2).

reorder :-
  retract(syntax(A,B,_)),
  A \= text(_),
  assertz(saved(A,B)),
  fail.
reorder :-
   saved(A,_),
   retract(saved(A,B)),
   assertz(syntax(A,B,both)),
   fail.
reorder.

term_expansion((A::=B),syntax(A,B,File)) :-
  loadingFrom(File).


term_expansion( (A->B;C), mapping(node,A,B,C) ) :- !.
term_expansion( (N+A->B), mapping(arg,N,A,B) ) :- !.
term_expansion( (A->B), mapping(void,A,B,void) ) :- !.
term_expansion( (A->B;C), mapping(node,A,B,C) ) :- !.


load(L) :-
  retractall(loadingFrom(_)),
  assert(loadingFrom(L)),
  consult(L).

go :-
  header,
  retractall(syntax(_,_,_)),
  load(lite),
  deleteFlags(lite),
  dbg,
  wlist(['<h2>OWL Lite Abstract Syntax</h2>',nl]),
  syn(lite),
  load(dl),
  wlist(['<h2>OWL DL Incremental Abstract Syntax</h2>',nl]),
  dbg,
  syn(dl),  dbg,
  deleteFlags(dl),
  reorder,
  deleteFlags(both),
  wlist(['<h2>OWL DL Complete Abstract Syntax</h2>',nl]),
  syn(both),
  deleteFlags(both).

dbg :-
  flag(description,N,N),
  wuser([N,nl]).
deleteFlags(Lvl) :-
  syntax(A,_,Lvl),
  atomic(A),
  flag(A,_,1),
  fail.
%deleteFlags(dl):-!.
deleteFlags(Lvl) :-
  syntax(A,_,Lvl),
  atomic(A),
  linkName(A,Lnk,Lvl),
  flag(Lnk,_,0),
  fail.
deleteFlags(_).


:-dynamic lastSymbol/1.
:-dynamic makeLink/2.
:-dynamic linked/1.
syn(Lvl) :-
  retractall(makeLink(_,_)),
  retractall(lastSymbol(_)),
  syntax(A,B,Lvl),
  writeAS(A,B,Lvl),
  fail.
syn(_) :- clearLastSymbol.

writeAS(text(Type),Text,Lvl) :-
  !,
  clearLastSymbol,
  write('<p>'),
  commentType(Type),
  wComment(Text,Lvl),
  write('</p>'),
  nl.

writeAS(Symbol,Rule,Lvl) :-
  lastSymbol(Symbol),
  !,
  letter(Symbol,L),
  flag(Symbol,Cnt,Cnt+1),
  wlist([sp(12),'[',L,Cnt,'] | ']),
  writeRuleBody(Rule,Lvl).


writeAS(Symbol,Rule,Lvl) :-
  clearLastSymbol,
  assert(lastSymbol(Symbol)),
  linkName(Symbol,Link,Lvl),
  flag(Link,Cnt,Cnt+1),
  letter(Symbol,L),
  flag(Symbol,LCnt,LCnt+1),
  level2lvlID(Lvl,LID),
  wlist([nl,'<pre><a name="',Link,'-',Cnt,'" id="',Link,'-',Cnt,'"></a>']),
  (Cnt=0->wlist(['<a name="',Link,'" id="',Link,'"></a>']);true),
  atom_length(Symbol,Lg),
  wlist([Symbol,sp(12-Lg),'<a name="',LID,L,LCnt,'" id="',LID,L,LCnt,'"></a>','[',L,LCnt,'] ::= ']),
  writeRuleBody(Rule,Lvl).

level2lvlID(both,'') :-!.
level2lvlID(_,'x-').
letter(axiom,'X') :- !.
letter(cardinality,'K') :- !.
letter(dataLiteral,'L') :- !.
letter(directive,'V') :- !.
letter(metaproperty,'Z') :- !.
letter(classInRestriction,'S') :- !.
letter(dataRange,'G') :- !.
letter(Symbol,L) :- 
   sub_atom(Symbol,0,1,_,Let),
   upcase_atom(Let,L).

clearLastSymbol :-
  retract(lastSymbol(_)),
  wlist(['</pre>',nl]),
  fail.
clearLastSymbol.

writeRuleBody(Rule,Lvl) :-
  formatRuleBody(Rule,Lvl,true,20,_),
  nl.

% formatRuleBody(Body,Lvl,Top,InCol,OutCol) 
formatRuleBody([Opt],Lvl,_Top) -->
    !,
    output('[ '),
    formatRuleBody(Opt,Lvl,fail),
    output(' ]').

formatRuleBody(+Terminal,_Lvl,_Top) -->
    !,
    {write('<span class="terminal">')},
    output(Terminal),
    {write('</span>')}.

formatRuleBody({Many},Lvl,_Top) -->
    !,
    output('{ '),
    formatRuleBody(Many,Lvl,fail),
    output(' }').


formatRuleBody(ID,_Lvl,_Top) -->
   {uriref(ID,Desc),
   !,
   wlist(['<a href="#uri-',Desc,'ID">'])},
   output(Desc),
   output('ID'),
   {write('</a>')}.

formatRuleBody(Expression,Lvl,Top) -->
   { Expression =.. [A,A1|T] },
   !,
   { sub_atom(A,0,1,_,First),
     upcase_atom(First,Upper),
     sub_atom(A,1,_,0,Rest) },
   output('\''),
   output(Upper),
   output(Rest),
   output('(\' '),
   formatRuleBodyAll([A1|T],Lvl,Top),
   output('\')\'').
   


formatRuleBody(NonTerminal,Lvl,_Top) -->
   {linkName(NonTerminal,Link,Lvl),
   !,
   wlist(['<a href="#',Link,'"><span class="nonterminal">'])},
   output(NonTerminal),
   {write('</span></a>')}.
       
formatRuleBodyAll([],_,_) --> [].
formatRuleBodyAll([A1|T],Lvl,Top) -->
   optNL(Top),
   formatRuleBody(A1,Lvl,fail),
   output(' '),
   formatRuleBodyAll(T,Lvl,Top).
   
optNL(true,N,25) :-
   N > 60,
   !,
   nl,
   write('                         ').
optNL(_) --> [].
   
   
output(W,N,N1) :-
   atom_length(W,L),
   N1 is N + L,
   write(W).


commentType(comment).
commentType(rationale) :-
   write('<strong>Rationale:</strong> ').

wComment([],_Lvl).
wComment([next(Item)|T],Lvl) :-
  !,
  flag(link,N,N+1),
  atom_concat(miscLink,N,Link),  
  assert(makeLink(Item,Link)),
  wlist(['<a href="#',Link,'">',Item,'</a>',nl]),
  wComment(T,Lvl).
wComment([link(ID)|T],Lvl) :-
   uriref(ID,Desc),
   !,
   wlist(['<a href="#uri-',Desc,'ID">',Desc,'ID</a>',nl]),
  wComment(T,Lvl).
wComment([link(Item)|T],Lvl) :-
  !,
  linkName(Item,Link,Lvl),
  wlist(['<a href="#',Link,'">',Item,'</a>',nl]),
  wComment(T,Lvl).
wComment([Other|T],Lvl) :-
   write(Other),nl,
   wComment(T,Lvl).

checkLevel(dl,dl).
checkLevel(_,lite).

abstract(Expression,Lvl) :-
  abstract(ontology,Expression,Lvl),
  (Lvl=lite->true;true).


abstract(Goal,Lvl) -->
  {syntax(Goal,_,_)},
  !,
  {syntax(Goal,Expand,Lvl1),
  checkLevel(Lvl,Lvl1)},
  abstract(Expand,Lvl).  
abstract(U,Lvl) -->
  { uriref(U,Functor) },
  !,
  [URI],
  { URI =.. [Functor,URIref],
    detailed(U,URIref,Lvl) }.


abstract({_},_) --> [].
abstract({Many},Lvl) --> !,
   abstractall([Many,{Many}],Lvl).
abstract([_],_) --> [].
abstract([Opt],Lvl) --> !,
   abstract(Opt,Lvl).

abstract(+Const,_) --> !,
   [Const].
abstract(Functor,Lvl) -->
  {Functor =.. [Hd|Args],
   Args = [_|_],
   expecting(Hd)},
   !,
   [Val],
   { Val =.. [Hd|VArgs],
     abstractall(Args,Lvl,VArgs,[]) }.
abstract(Other,_) -->
   { functor(Other,O,N),
   throw(unrecognized(O/N))}.

expecting(ontology).
expecting(meta).
expecting(header).


abstractall([A|B],Lvl) -->
  abstract(A,Lvl),
  abstractall(B,Lvl).
abstractallall([],_) --> [].


detailed(m,URIref,_) :- !,
   URIref = owl:Sp,
   member(Sp,[imports,backwardsCompatibleWith]).
detailed(dr,URIref,Lvl) :- !,
   (URIref = rdfs:'Literal';Lvl=dl),
   !.
detailed(_,_,_).

linkName(Item,Link,lite) :-
   !,
   
   (syntax(Item,_,lite)->
     atom_concat('lite-syntax-',Item,Link);
   throw(undefinedTerm(Item,lite))).
linkName(Item,Link,Lin) :-
   syntax(Item,_,Lvl)->
     atom_concat('-syntax-',Item,LinkX),
     atom_concat(Lvl,LinkX,Link);
   throw(undefinedTerm(Item,Lin)).

   
wlist([nl|T]) :- !, nl, wlist(T).
wlist([sp(N)|T]) :-
   !,
   N1 is N,
   spaces(N1),
   wlist(T).
wlist([X|T]) :- write(X), wlist(T).
wlist([]).

spaces(N) :- N < 1, !.
spaces(N) :- put(" "),N1 is N - 1, spaces(N1).
wuser(L) :-
   telling(T),
   tell(user),
   wlist(L),
   tell(T).


header :- wlist( ['<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
         "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html lang="en-US" xml:lang="en-US" xmlns="http://www.w3.org/1999/xhtml">',

'<head>
<meta http-equiv="Content-Type" content="text/html; charset=ISO-8859-1" />
<title>AS reconsidered</title>',


'<link rel="stylesheet" type="text/css"
      href="http://www.w3.org/StyleSheets/TR/base" />'

,'</head>
<body>'] ).



%--------------
% Mapping Rules
%--------------

map :-
  flag(mapno,_,1),
  write('<table border="1" width="100%">
<tr>
<th></th><th>Input</th><th>Triples</th><th>Return</th><th>Notes</th>
</tr>'),
  mapping(V,In,Triples,Out),
  (V=void;V=node),
  In \= text2,
  lmap(V,In,Triples,Out),
  fail.

map :-
  mapping(V,In,Triples,Out),
  (In=text2;V=arg),
  lmap(V,In,Triples,Out),
  fail.

map :- write('</table>
').
:- dynamic note/1.
rhs2list(true,[]) :- !.
rhs2list((A,B),[A|T]) :-
  !,
  rhs2list(B,T).
rhs2list(A,[A]).

head2ID(Hd,Id) :-
  name(Hd,N),
  maplist(space2minus,N,NN),
  append("maprules-",NN,NNN),
  name(Id,NNN).

space2minus(32,45) :- !.
space2minus(X,X).

text(text).
text(text2).

lmap(_,_,_,_) :-
  retractall(note(_)),
  fail.
lmap(void,TEXT,heading(1,Hd),void) :-
  text(TEXT),
  !,
  head2ID(Hd,ID),
  wlist(['<tr><td colspan="5" class="h1"><a name="',ID,'" id="',ID,
         '"><span class="h1">',Hd,'</span></a></td></tr>',nl]).

lmap(void,TEXT,heading(2,Hd),void) :-
  text(TEXT),
  !,
  head2ID(Hd,ID),
  wlist(['<tr><td></td><td colspan="4" class="h2"><a name="',ID,'" id="',ID,
         '"><span class="h2">',Hd,'</span></td></tr>',nl]).

lmap(void,TEXT,Atom,void) :-
  text(TEXT),
  !,
  wlist(['<tr><td colspan="5">',Atom,'</td>',nl]).
lmap(node,Functor,RHS,Result) :-
  Functor =.. [U,Arg],
  uriref(U,URIType),
  !,
  flag(mapno,Cnt,Cnt+1),
  writeRuleNumber(Cnt),
  uriarg(URIType,Arg,Functor),
  rhs2list(RHS,RL),
  simpleRHS(RL,XL),
  write('<td>'),
  writeSimpleRHS(XL),
  write('</td>'),nl,
  writeReturn(Result,1),
  writeNotes.

lmap(node,A,RHS,Result) :-
  atomic(A),
  !,
  flag(mapno,Cnt,Cnt+1),
  writeRuleNumber(Cnt),
  wlist(['<td><code>',A,'</code></A></td>',nl]),
  rhs2list(RHS,RL),
  simpleRHS(RL,XL),
  write('<td>'),
  writeSimpleRHS(XL),
  write('</td>'),nl,
  writeReturn(Result,1),
  writeNotes.
  

lmap(NA,Functor,RHS,Result) :-
  (NA = void;NA = node),
  !,
  Functor =.. [Construct|Parts],
  flag(mapno,Cnt,Cnt+1),
  flag(submapno,_,1),
  rhs2list(RHS,RL),
  expandLeft(Parts,LL),
  convertFunctor(Construct,Open),
  append([Open,indent|LL],[noindent,')'],LLL),
  lmapall(LLL,RL,SubRules,Result),
  flag(indentation,_,0),
  length(SubRules,Ln),
  SubRules = [SR1|SRT],
  writeComplexResult(SR1,Ln),
  checklist(writeComplex,SRT).

lmap(arg,'<strong>*</strong>',F,RHS) :-
  !,
  lmap(void,F,RHS,void).

writeRuleNumber(Cnt) :-
  wlist(['<tr><td><a name="map-',Cnt,'" id ="map-',Cnt,
          '"></a><a href="#map-',Cnt,'">',
          Cnt,'</a></td>',nl]).

writePairRuleNumber :-
  flag(mapno,M,M),
  flag(submapno,S,S+1),
  concat_atom([M,'.',S],NN),
  writeRuleNumber(NN).

% writeComplex(SubRule)

writeComplex(X) :-
  writeComplexResult(X,0).

writeComplexResult({Star},Ln) :-
  !,
  writePairRuleNumber,
  writeComplex('{ ',' }',Star,Ln).

writeComplexResult([Star],Ln) :-
  !,
  writePairRuleNumber,
  writeComplex('[ ',' ]',Star,Ln).

writeComplexResult(+sub(L,R,N,X),Ln) :-
  !,
  writePairRuleNumber,
 write('<td>'),
 indent,
 write('{ '),
 writeLeft(L),
 write(' }</td>'),nl,
 write('<td>'),
 writeSimpleRHS(R),
 write('</td>'),nl,
 writeReturn(X,Ln),
 writeNotes(N).
  

writeComplexResult(sub(L,[],[],void),Ln) :-
  !,
  flag(submapno,Cnt,Cnt),
  (Cnt=1->
       flag(mapno,N,N),
       writeRuleNumber(N);
     write('<tr><td></td>')
  ),
  writeComplex('','',sub(L,[],[],void),Ln).
  

writeComplexResult(Sub,Ln) :-
  !,
  writePairRuleNumber,
  writeComplex('','',Sub,Ln).

writeComplex(Left,Right,sub(L,R,N,X),Ln) :-
 write('<td>'),
 indent,
 write(Left),
 writeLeft(L),
 write(Right),
 write('</td>'),nl,
 write('<td>'),
 write(Left),
 writeSimpleRHS(R),
 write(Right),
 write('</td>'),nl,
 writeReturn(X,Ln),
 writeNotes(N).
 
indent :- 
  flag(indentation,Cnt,Cnt),
  length(L,Cnt),
  checklist(nbsp,L).
nbsp(_) :- write('&nbsp;').

  

%expandLeft
expandLeft([],[]).
expandLeft([X|T],[X|TT]) :-
  noexpand(X),
  !,
  expandLeft(T,TT).
expandLeft([Fun|T],Exp) :-
  Fun =.. [F|Args],
  expandLeft(Args,AA),
expandLeft(T,TT),
  convertFunctor(F,Open),
  append([Open,indent|AA],[noindent,')'|TT],Exp).

convertFunctor(F,Open) :-
   sub_atom(F,1,_,0,Rest),
   sub_atom(F,0,1,_,First),
   upcase_atom(First,Up),
   concat_atom([Up,Rest,'('],Open).
  

noexpand({_}).
noexpand([_]).
noexpand(_-_).
noexpand(+_).
noexpand(A) :- atomic(A).

% lmapall(Parts,RHS,SubRules).

lmapall([],[],[],Result) :- !.
lmapall([{Lstar}|LT],[{RStar}|RT],[{sub([Lstar],XL,Notes,Result)}|TT],Result) :-
  !,
  rhs2list(RStar,RL),
  simpleRHS(RL,XL),
  collectNotes(Notes),
  lmapall(LT,RT,TT,void).
lmapall([{Lstar}|LT],[t(A,B,x(seq({RStar})))|RT],[+sub([Lstar],XL,Notes,Result)|TT],Result) :-
  !,
  simpleRHS([t(A,B,x(seq({RStar})))],XL),
  collectNotes(Notes),
  lmapall(LT,RT,TT,void).
lmapall([[Lstar]|LT],[[RStar]|RT],[[sub([Lstar],XL,Notes,Result)]|TT],Result) :-
  !,
  rhs2list(RStar,RL),
  simpleRHS(RL,XL),
  collectNotes(Notes),
  lmapall(LT,RT,TT,void).
lmapall(L,R,  [sub(LP,XL,Notes,Result)|TT],Result) :-
  nonOptionalPrefix(LP,L,LT),
  nonOptionalPrefix(RP,R,RT),
  simpleRHS(RP,XL),
  collectNotes(Notes),
  lmapall(LT,RT,TT,void).
nonOptionalPrefix([],[],[]) :- !.
nonOptionalPrefix([],[[X]|T],[[X]|T]):-!.
nonOptionalPrefix([],[{X}|T],[{X}|T]):-!.
nonOptionalPrefix([],[t(A,B,x(seq({S})))|T],[t(A,B,x(seq({S})))|T]):-!.
nonOptionalPrefix([H|T],[H|TT],R) :-
  nonOptionalPrefix(T,TT,R).
  
  

writeSimpleRHS([]).
writeSimpleRHS([X]) :-
  !,
  wlist(X).
writeSimpleRHS([H|T]) :-
  wlist(H),
  wlist(['<br />',nl]),
  writeSimpleRHS(T).

writeLeft([]) :- !.
writeLeft([H|T]) :-
  writeLeft1(H),
  writeLeft(T).

writeLeft1(indent) :- !,
  flag(indentation,I,I+3).
writeLeft1(noindent) :- !,
  flag(indentation,I,I-3).
writeLeft1(X) :-
  outputFixup(X,XX),
  wlist(XX),put(32).

writeReturn(_,0) :- !.
writeReturn(void,N) :-
  !,
  wlist(['<td rowspan="',N,'"></td>']).
writeReturn(Result,N) :-
  outputFixup(Result,R),
  wlist(['<td valign="top" rowspan="',N,'">'|R]),
  wlist(['</td>',nl]).

collectNotes(Notes) :-
  bagof(['<a href="#note-',N,'">',N,'</a>'],retract(note(N)),Notes),
  !.
collectNotes([]).
writeNotes :-
  collectNotes(Notes),
  writeNotes(Notes).
writeNotes(Notes) :-
  write('<td>'),
  writeSimpleRHS(Notes),
  write('</td></tr>'),
  nl.
  
  

subscript(A-N,A,['<sub>',N,'</sub>']) :- !.
subscript(Any,Any,[]).
outputFixup('<strong>*</strong>','<code><strong>@</strong></code>') :- !.
outputFixup('<em>uriref</em>','<em>uriref</em>') :- !.
outputFixup(blank,['<em>_:b</em>']) :-!.
outputFixup(x(U),Out) :-
   subscript(U,U1,Subscript),
   uriref(U1,URI),
   !,
   flatten( [
   [
    '<span class="transform">x</span>( <code>',URI,'ID</code>( <em>' , U1 ],
   Subscript,
   ['</em> ) )'] ], Out ).
outputFixup(A^^X:D,Out) :-
   !,
   outputFixup(A^^ (X:D),Out).
outputFixup(A@L^^D,Out) :-
   !,
   outputFixup(D,DT),
   Out = ['<code>',A,'@',L,'^^</code>'|DT].
outputFixup(A^^D,Out) :-
   !,
   outputFixup(D,DT),
   Out = ['<code>',A,'^^</code>'|DT].
outputFixup(A@L,Out) :-
   !,
   Out = ['<code>',A,'@',L,'</code>'].
outputFixup({A},Out) :-
   !,
   outputFixup(A,AA),
   append(['{ '|AA],[' }'],Out).
outputFixup(x(seq(A)),Out) :-
   !,
   outputFixup(seq(A),AA),
   append(
   [
    '<span class="transform">x</span>( ' |
   AA],[' )'],Out).
outputFixup(x(NT0),Out) :-
   !,
   subscript(NT0,NT,Subscript),
   name(NT,_), % checks NT is atomic
   append(
   [
    '<span class="transform">x</span>( <code>',NT |
   Subscript],['</code> )'],Out).
outputFixup(x(In,NT0),Out) :-
   !,
   subscript(NT0,NT,Subscript),
   name(NT,_), % checks NT is atomic
   outputFixup(In,InOut),
   flatten([
   [
    '<span class="transform">y</span>( '],
   InOut,
   [ ', <code>',NT ],
   Subscript,
   ['</code> )']
   ],Out).
outputFixup(U1,Out) :-
   U1 =.. [U,A],
   uriref(U,URItype),
   !,
   outputFixup(A,AO),
   flatten([
     ['<code>',URItype,'ID</code>( '],
     AO,
     [' )']
     ], Out ).
outputFixup(U,Out) :-
   subscript(U,U1,Subscript),
   uriref(U1,URItype),
   !,
   flatten([
     ['<code>',URItype,'ID</code>( <em>' , U1 ],
   Subscript,
   ['</em> )' ]
     ], Out ).
outputFixup(+X,['<em>',XX,'</em>']) :-
    !,
    capitalize(X,XX).
outputFixup(Namespace:'<em>built-in</em>',
     ['<code><strong>',Namespace,':</strong></code><em>built-in</em>']) :- !.
outputFixup(Namespace:Frag,['<code><strong>',Namespace,':',FixedFrag,'</strong></code>']) :-
     !,
     caseAdjustBuiltin(Namespace,Frag,FixedFrag).
outputFixup(Functor,XX) :-
  Functor =.. [A,R|T],
  A \= (-),
  !,
  capitalize(A,F),
  maplist(outputFixupSpace,[R|T],FF),
  flatten([
    ['<code>',F,'</code>( '],
    FF,
    [')']
    ],XX).
  
outputFixup(X,['<code>',X1,'</code>'|Sub]) :- 
     subscript(X,X1,Sub).
outputFixupSpace(X,Lsp) :-
   outputFixup(X,L),
   append(L,[' '],Lsp).

% needs to check that F is in Namespace, possibly upcasing first letter.
caseAdjustBuiltin(Namespace,F,F) :-
   builtin(Namespace,F), !.
caseAdjustBuiltin(Namespace,F,Upcased) :-
   capitalize(F,Upcased),
   builtin(Namespace,Upcased), !.
caseAdjustBuiltin(Namespace,F,_) :-
   throw(builtin(Namespace,F)).

capitalize(F,Upcased) :-
   sub_atom(F,1,_,0,Rest),
   sub_atom(F,0,1,_,First),
   upcase_atom(First,Up),
   atom_concat(Up,Rest,Upcased).
  
addNote(X) :-
  assert(note(X)).

uriarg(URIType,Arg,Combo) :-
  binduriarg(URIType,Arg),
  outputFixup(Combo,AOut),
  wlist(['<td>'|AOut]),
  wlist(['</td>',nl]).
binduriarg(_URIType,'<em>uriref</em>') :- !.
binduriarg(URIType,_Namespace:'<em>built-in</em>') :-
   !,
   addNote(URIType).
binduriarg(_,_).

% simpleRHS( +RHSList, ?RHSTriples )
simpleRHS( A, B ) :-
  maplist(lineRHS,A,B).
/*
simpleRHS( [+[Opt]|T], [ OptOut2 | TT ] ) :-
   !,
   simpleRHS( [Opt|T], [ OptOut1 | TT ] ),
   append(['[ '|OptOut1],[' ] (opt)'],OptOut2 ).
simpleRHS([note(N)/X|T],[XX|TT]) :-
   !,
   addNote(N),
   simpleRHS([X|T],[XX|TT]).
simpleRHS([t(Subj,Pred,Obj)|T],[Out|TT]) :-
   !,
   outputFixup(Subj,S),
   outputFixup(Pred,P),
   outputFixup(Obj,O),
   flatten([S,['&nbsp;'],P,['&nbsp;'],O,['&nbsp;.']],Out),
   simpleRHS(T,TT).
simpleRHS([x(Part)|T],[XO|TT]) :-
   !,
   outputFixup(x(Part),XO),
   simpleRHS(T,TT).
simpleRHS([x(Arg,Part)|T],[XO|TT]) :-
   !,
   outputFixup(x(Arg,Part),XO),
   simpleRHS(T,TT).
simpleRHS([Bad|_],_) :-
   throw(badMapRule(Bad)).
*/

%%%%

lineRHS( +[Opt],  OptOut2  ) :-
   !,
   lineRHS( Opt,  OptOut1  ),
   append(['[ '|OptOut1],[' ] (opt)'],OptOut2 ).
lineRHS(note(N)/X,XX) :-
   !,
   addNote(N),
   lineRHS(X,XX).
lineRHS(t(Subj,Pred,Obj),Out) :-
   !,
   outputFixup(Subj,S),
   outputFixup(Pred,P),
   outputFixup(Obj,O),
   flatten([S,['&nbsp;'],P,['&nbsp;'],O,['&nbsp;.']],Out).
lineRHS(x(Part),XO) :-
   !,
   outputFixup(x(Part),XO).
lineRHS(x(Arg,Part),XO) :-
   !,
   outputFixup(x(Arg,Part),XO).
lineRHS(Bad,_) :-
   throw(badMapRule(Bad)).











